/*
 * Copyright (C) 2010 TimeSys Corporation
 */
#include <stdio.h>
#include <fcntl.h>

#include <linux/fb.h>
#include <linux/videodev2.h>

#include <gst/gst.h>
#include <gst/interfaces/tuner.h>


/* Board-specific default and commented out alternate options */
#if defined(BOARD_dm365_evm)

#define CAPTURE_SUPPORTED
#define DEFAULT_VIDEO_FILENAME "sample.m4v"
#define DEFAULT_VIDEO_ENCODER "mpeg4enc"
#define DEFAULT_VIDEO_STANDARD "D1_NTSC"
#define DEFAULT_VIDEO_OUT "COMPOSITE"
#define DEFAULT_VIDEO_INPUT "Composite"
//#define DEFAULT_VIDEO_INPUT "S-Video"

/* Settings for progressive scan/high def */
/* XXX: beware, 720P capture via component is not working via gstreamer. */
//#define DEFAULT_VIDEO_STANDARD "480P"
//#define DEFAULT_VIDEO_STANDARD "720P_60"
//#define DEFAULT_VIDEO_OUT "COMPONENT"
//#define DEFAULT_VIDEO_INPUT "Component"

/* Camera is untested */
//#define DEFAULT_VIDEO_INPUT "Camera"

#elif defined(BOARD_omap3_evm)

#define DEFAULT_VIDEO_FILENAME "sample.m4v"
#define DEFAULT_VIDEO_STANDARD "VGA"
#define DEFAULT_VIDEO_OUT "LCD"
#define OUTPUT_ROTATION 90

//#define DEFAULT_VIDEO_STANDARD "480P"
//#define DEFAULT_VIDEO_OUT "DVI"
#else

#error "Unknown BOARD"

#endif

/* settings:
 * TODO: eventually it would be nice to have these settable at runtime */

char *filename = DEFAULT_VIDEO_FILENAME;
char *displayVideoStd = DEFAULT_VIDEO_STANDARD;
char *displayVideoOut = DEFAULT_VIDEO_OUT;
#ifdef CAPTURE_SUPPORTED
char *encoder_name = DEFAULT_VIDEO_ENCODER;
char *captureVideoInput = DEFAULT_VIDEO_INPUT;
#endif

static int quit;

char do_read() {
  char ret = 0, next;
  do {
    next = getchar();
    if (ret == 0) {
      ret = next;
    } else if (ret == 1) {
      /* do nothing */
    } else if (next == '\n') {
      break;
    } else {
      puts("Multi-char commands not allowed");
      ret = 1;
    }
  } while (next != '\n');
  return ret;
}

static gboolean bus_callback(GstBus *bus, GstMessage *message, gpointer data) {
  g_print ("Got %s message\n", GST_MESSAGE_TYPE_NAME (message));
  switch (GST_MESSAGE_TYPE (message)) {
    case GST_MESSAGE_ERROR: {
      GError *err;
      gchar *debug;

      gst_message_parse_error (message, &err, &debug);
      g_print ("Error: %s\n", err->message);
      g_error_free (err);
      g_free (debug);

      quit = TRUE;
      break;
    }
    case GST_MESSAGE_EOS:
      /* end-of-stream */
      quit = TRUE;
      break;
    default:
      /* unhandled message */
      break;
  }

  /* we want to be notified again the next time there is a message
   * on the bus, so returning TRUE (FALSE means we want to stop watching
   * for messages on the bus and our callback should not be called again)
   */
  return TRUE;
}

void pauseToggle(GstElement *pipeline) {
  GstState state;
  GstStateChangeReturn ret;

  ret = gst_element_get_state(pipeline, &state, NULL, 0);
  if (ret == GST_STATE_CHANGE_FAILURE) {
    puts("Cannot determine pipeline state");
    printf("return value: %u\n", ret);
    return;
  }
  switch (state) {
  case GST_STATE_PAUSED:
    puts("Resuming pipleline");
    gst_element_set_state(pipeline, GST_STATE_PLAYING);
    break;
  case GST_STATE_PLAYING:
    puts("Pausing pipeline");
    gst_element_set_state(pipeline, GST_STATE_PAUSED);
    break;
  default:
    printf("Cannot Pause/Resume in pipeline state: %s\n", 
	gst_element_state_get_name(state));
  }
}

void printState(GstElement *pipeline) {
  GstState state;
  GstStateChangeReturn ret;

  ret = gst_element_get_state(pipeline, &state, NULL, 0);
  if (ret == GST_STATE_CHANGE_FAILURE) {
    puts("Cannot determine pipeline state");
    return;
  }
  printf("Current pipeline state: %s\n", 
	gst_element_state_get_name(state));
}

void checkBus(GstElement *pipeline) {
  GstMessage *message;
  GstBus *bus;

  bus = gst_pipeline_get_bus(GST_PIPELINE(pipeline));
  while ((message = gst_bus_pop(bus)) != NULL) {
    bus_callback(bus, message, NULL);
    gst_message_unref(message);
  }
  gst_object_unref(bus);
}

void pipeline_menu(GstElement *pipeline) {
  int response;

  quit = 0;
  do {
    puts("[P] - Play/Pause");
    puts("[S] - Check pipeline state and events");
    puts("[Q] - Quit");
    fflush(stdout);
    response = do_read();
    switch (response) {
    case 'p':
    case 'P':
      pauseToggle(pipeline);
      break;
    case 's':
    case 'S':
      checkBus(pipeline);
      printState(pipeline);
      break;
    case 'Q':
    case 'q':
      quit = TRUE;
      break;
    default:
      puts("Invalid command");
    }
  } while (!quit);
}

void start_play(void) {
  GstElement *pipeline, *src, *typefind, *decoder, *sink;

  printf("Playing movie: %s\n", filename);

  pipeline = gst_pipeline_new("play_pipeline");

  src = gst_element_factory_make("filesrc", "filesrc");
  g_object_set(G_OBJECT(src),
	"location", filename,
	NULL);

  /* These two together will automatically detect the type and apply
   * the correct TI codec.
   */
  typefind = gst_element_factory_make("typefind", "typefind");
  decoder = gst_element_factory_make("TIViddec2", "decoder");

  sink = gst_element_factory_make("TIDmaiVideoSink", "video_sink");
  g_object_set(G_OBJECT(sink),
	"videoStd", displayVideoStd,
	"videoOutput", displayVideoOut,
#if defined(OUTPUT_ROTATION)
	"rotation", OUTPUT_ROTATION,
#endif
	"sync", FALSE,
	NULL);
  
  gst_bin_add_many(GST_BIN(pipeline), src, typefind, decoder, sink, NULL);

  if (!gst_element_link_many(src, typefind, decoder, sink, NULL)) {
    g_warning("Failed to link elements!");
  }

  gst_element_set_state (pipeline, GST_STATE_PLAYING);

  pipeline_menu(pipeline);
  checkBus(pipeline);
  /* clean up */
  gst_element_set_state (pipeline, GST_STATE_NULL);
  gst_object_unref (pipeline);

}

#ifdef CAPTURE_SUPPORTED
void start_capture(void) {
  GstElement *pipeline, *src, *encoder, *sink, *tee, *queue1, *queue2, *monitor;
  GstTunerChannel *channel;
  int videofd;
  v4l2_std_id standard = 0;

  printf("Capturing to file: %s (%s)\n", filename, encoder_name);

  /* TODO: Display video on-screen as it is being captured.
   *   Using a gstreamer "tee" to display during capture causes
   *   instability in the display driver.  Using the hardware
   *   previewer may be a better option.
   */
  pipeline = gst_pipeline_new("capture_pipeline");

  src = gst_element_factory_make("v4l2src", "capturesrc");
  g_object_set(G_OBJECT(src),
	"always-copy", FALSE,
	NULL);

  encoder = gst_element_factory_make("TIVidenc1", "encoder");
  g_object_set(G_OBJECT(encoder),
        "engineName", "encode",
        "codecName", encoder_name,
        NULL);

  sink = gst_element_factory_make("filesink", "file_sink");
  g_object_set(G_OBJECT(sink),
	"location", filename,
	NULL);

  gst_bin_add_many(GST_BIN(pipeline), src, encoder, sink, NULL);

  if (!gst_element_link_many(src, encoder, sink, NULL)) {
    g_warning("Failed to link elements!");
  }

  gst_element_set_state(pipeline, GST_STATE_READY);

  /* Input srcs (composite, component, s-vid, etc.) are modeled as 
   * channels in gstreamer.
   */
  channel = gst_tuner_find_channel_by_name(GST_TUNER(src), captureVideoInput);
  gst_tuner_set_channel(GST_TUNER(src), channel);

  /* Setting the norm through gstreamer doesn't seem to work.
   * Automatically detect video std with v4l2 calls. */
  g_object_get(G_OBJECT(src), "device-fd", &videofd, NULL);
  if (ioctl(videofd, VIDIOC_QUERYSTD, &standard) < 0) {
    perror("Query standard");
  } else {
    if (ioctl(videofd, VIDIOC_S_STD, &standard) < 0) {
      perror("Set standard");
    }
  }

  gst_element_set_state (pipeline, GST_STATE_PLAYING);

  printf("Capture is started from %s\n", captureVideoInput);
  pipeline_menu(pipeline);
  checkBus(pipeline);
  /* clean up */
  gst_element_set_state (pipeline, GST_STATE_NULL);
  gst_object_unref (pipeline);
}

void start_loopback(void) {
  GstElement *pipeline, *src, *encoder, *typefind, *decoder, *sink;
  GstTunerChannel *channel;
  GstState state;
  int videofd;
  v4l2_std_id standard = 0;

  printf("Loopback video\n");

  pipeline = gst_pipeline_new("loopback_pipeline");

  src = gst_element_factory_make("v4l2src", "capturesrc");
  g_object_set(G_OBJECT(src),
	"always-copy", FALSE,
	NULL);


  sink = gst_element_factory_make("TIDmaiVideoSink", "video_sink");
  g_object_set(G_OBJECT(sink),
	"videoStd", displayVideoStd,
	"videoOutput", displayVideoOut,
#if defined(OUTPUT_ROTATION)
	"rotation", OUTPUT_ROTATION,
#endif
	NULL);
   
  
  gst_bin_add_many(GST_BIN(pipeline), src, sink, NULL);

  if (!gst_element_link_many(src, sink, NULL)) {
    g_warning("Failed to link elements!");
  }

  gst_element_set_state(pipeline, GST_STATE_READY);
  /* Input srcs (composite, component, s-vid, etc.) are modeled as 
   * channels in gstreamer.
   */
  channel = gst_tuner_find_channel_by_name(GST_TUNER(src), captureVideoInput);
  gst_tuner_set_channel(GST_TUNER(src), channel);

  /* Setting the norm through gstreamer doesn't seem to work.
   * Automatically detect video std with v4l2 calls. */
  g_object_get(G_OBJECT(src), "device-fd", &videofd, NULL);
  if (ioctl(videofd, VIDIOC_QUERYSTD, &standard) < 0) {
    perror("Query standard");
  } else {
    if (ioctl(videofd, VIDIOC_S_STD, &standard) < 0) {
      perror("Set standard");
    }
  }

  gst_element_set_state (pipeline, GST_STATE_PLAYING);

  pipeline_menu(pipeline);
  checkBus(pipeline);
  /* clean up */
  gst_element_set_state (pipeline, GST_STATE_NULL);
  gst_object_unref (pipeline);
}
#endif

char main_menu(void) {
  puts("[P] - Play File");
#ifdef CAPTURE_SUPPORTED
  puts("[C] - Capture To File");
  puts("[L] - Loopback Video (Capture and Display)");
#endif
  /*puts("[S] - Settings");*/
  puts("[Q] - Quit");
  fflush(stdout);
  return do_read();
}


#if defined(BOARD_dm365_evm)

#define OSD_DEVICE "/dev/fb2"
void hide_osd(void) {
  int fd, result, val=0;
  struct fb_var_screeninfo info;
  /* This assume OSD attribute window is /dev/fb2 */
  fd = open(OSD_DEVICE, O_WRONLY);
  if (fd < 0) {
    printf("Could not open osd attribute device: %s\n", OSD_DEVICE);
    perror("open");
    return;
  }

  if (ioctl(fd, FBIOGET_VSCREENINFO, &info) != 0) {
    perror("get_screeninfo");
    return;
  }

  if (! (info.nonstd != 0 && info.bits_per_pixel == 4)) {
    printf("Not an osd attribute window: %s\n", OSD_DEVICE);
    exit(-1);
  }
  for (;;) {
    result = write(fd, &val, 1);
    if (result < 0) {
      if (errno == ENOSPC) {
        break;
      }
      perror("write");
      break;
    }
  }
  close(fd);
}
#else // BOARD == dm365

#define hide_osd() {if (0) puts("Ignored");}

#endif

int main(int argc, char **argv) {
  char response;
  pid_t pid;

  /* Hide the OSD it obscures the video window */
  hide_osd();

  /* Initialize gstreamer */
  gst_init(&argc, &argv);

  for (;;) {
    response = main_menu();
    /* TI gstreamer plugins fail if they are used more than once in a single
     * process.  So we always fork before starting any pipelines.
     */
    if (response == 'Q' || response == 'q') {
      // need to check for quit before forking.
      break;
    }
    pid = fork();
    if (pid == -1) {
      perror("fork");
      continue;
    } else if (pid != 0) {
      /* I'm the parent, wait for the child. */
      waitpid(pid, 0, NULL);
      continue;
    }

    /* Only the child will reach this point. */
    switch (response) {
    case 'p':
    case 'P':
      start_play();
      break;
#ifdef CAPTURE_SUPPORTED
    case 'c':
    case 'C':
      start_capture();
      break;
    case 'l':
    case 'L':
      start_loopback();
      break;
#endif
    default:
      puts("Invalid choice");
    }
    puts("\n");
    /* Only the child will reach here and exit. */
    exit(0);
  }
}


